@@ -1,6 +1,4 @@
 /*
- * $Id: wacom.c,v 1.8 2006/05/05 17:25:02 pingc Exp $
- *
  *  Copyright (c) 2000-2002 Vojtech Pavlik  <vojtech@suse.cz>
  *  Copyright (c) 2000 Andreas Bach Aaen    <abach@stofanet.dk>
  *  Copyright (c) 2000 Clifford Wolf        <clifford@clifford.at>
@@ -57,6 +55,8 @@
  *    v2.4.30-pc0.7.0   - Added Volito2 support
  *    v2.4.30-pc0.7.1   - Added DTF720, DTU 710, G4
  *    v2.4.30-pc0.7.3   - Added DTF 521, I3 12x12, and I3 12x19
+ *    v2.4.30-pc0.7.5   - Support tablet buttons/keys  
+ *    v2.4.30-pc0.7.7   - Support Intuos outbound tracking
  */
 
 /*
@@ -87,7 +87,7 @@
 /*
  * Version Information
  */
-#define DRIVER_VERSION "v2.4.30-pc0.7.3"
+#define DRIVER_VERSION "v2.4.30-pc0.7.7"
 #define DRIVER_AUTHOR "Vojtech Pavlik <vojtech@suse.cz>"
 #ifndef __JEJ_DEBUG
 #define DRIVER_DESC "USB Wacom Graphire and Wacom Intuos tablet driver (LINUXWACOM)"
@@ -103,6 +103,7 @@ MODULE_LICENSE("GPL");
 #define STYLUS_DEVICE_ID	0x02
 #define CURSOR_DEVICE_ID	0x06
 #define ERASER_DEVICE_ID	0x0A
+#define ERASER_DEVICE_ID	0x0A
 
 struct wacom_features {
 	char *name;
@@ -267,8 +268,8 @@ static void wacom_penpartner_irq(struct urb *urb)
 				wacom->id[0] = (data[5] & 0x20) ? ERASER_DEVICE_ID : STYLUS_DEVICE_ID;
 				input_report_key(dev, wacom->tool[0], 1);
 				input_report_abs(dev, ABS_MISC, wacom->id[0]); /* report tool id */
-				input_report_abs(dev, ABS_X, data[2] << 8 | data[1]));
-				input_report_abs(dev, ABS_Y, data[4] << 8 | data[3]));
+				input_report_abs(dev, ABS_X, data[2] << 8 | data[1]);
+				input_report_abs(dev, ABS_Y, data[4] << 8 | data[3]);
 				input_report_abs(dev, ABS_PRESSURE, (signed char)data[6] + 127);
 				input_report_key(dev, BTN_TOUCH, ((signed char)data[6] > -127));
 				input_report_key(dev, BTN_STYLUS, (data[5] & 0x40));
@@ -306,7 +307,8 @@ static void wacom_graphire_irq(struct urb *urb)
 
 	if (urb->status) return;
 
-	if (data[0] == 99) return; /* for Volito tablets */
+	/* 2 Volito1 and 2 users say this is wrong
+	if (data[0] == 99) return; */
 
 	if (data[0] != 2)
 	{
@@ -315,7 +317,7 @@ static void wacom_graphire_irq(struct urb *urb)
 	}
 
 	id = STYLUS_DEVICE_ID;
-	if ( data[1] & 0x10 ) /* in prox */
+	if ( data[1] & 0x80 ) /* in prox */
 	{
 		switch ((data[1] >> 5) & 3) {
 
@@ -348,9 +350,6 @@ static void wacom_graphire_irq(struct urb *urb)
 					input_report_abs(dev, ABS_DISTANCE, data[7]);
 				break;
 		}
-	}
-
-	if (data[1] & 0x90) {
 		x = data[2] | ((__u32)data[3] << 8);
 		y = data[4] | ((__u32)data[5] << 8);
 		input_report_abs(dev, ABS_X, x);
@@ -362,12 +361,24 @@ static void wacom_graphire_irq(struct urb *urb)
 			input_report_key(dev, BTN_STYLUS, data[1] & 0x02);
 			input_report_key(dev, BTN_STYLUS2, data[1] & 0x04);
 		}
-	}
-	if (data[1] & 0x10)
 		input_report_abs(dev, ABS_MISC, id); /* report tool id */
-	else 
+		input_report_key(dev, wacom->tool[0], 1);
+	} else if (!(data[1] & 0x90)) {
+		input_report_abs(dev, ABS_X, 0);
+		input_report_abs(dev, ABS_Y, 0);
+		if (wacom->tool[0] == BTN_TOOL_MOUSE) {
+			input_report_key(dev, BTN_LEFT, 0);
+			input_report_key(dev, BTN_RIGHT, 0);
+			input_report_abs(dev, ABS_DISTANCE, 0);
+		} else {
+			input_report_abs(dev, ABS_PRESSURE, 0);
+			input_report_key(dev, BTN_TOUCH, 0);
+			input_report_key(dev, BTN_STYLUS, 0);
+			input_report_key(dev, BTN_STYLUS2, 0);
+		}
 		input_report_abs(dev, ABS_MISC, 0); /* reset tool id */
-	input_report_key(dev, wacom->tool[0], data[1] & 0x10);
+		input_report_key(dev, wacom->tool[0], 0);
+	}
 	input_event(dev, EV_MSC, MSC_SERIAL, data[1] & 0x01);
 
 	/* send pad data */
@@ -382,10 +393,12 @@ static void wacom_graphire_irq(struct urb *urb)
 			rw = ((data[7] & 0x18) >> 3) - ((data[7] & 0x20) >> 3);
 			input_report_rel(dev, REL_WHEEL, rw);
 			input_report_key(dev, BTN_TOOL_FINGER, 0xf0);
+			input_report_abs(dev, ABS_MISC, PAD_DEVICE_ID);
 			input_event(dev, EV_MSC, MSC_SERIAL, 0xf0);
 		} else if ( wacom->id[1] ) {
 			wacom->id[1] = 0;
 			input_report_key(dev, BTN_TOOL_FINGER, 0);
+			input_report_abs(dev, ABS_MISC, PAD_DEVICE_ID);
 			input_event(dev, EV_MSC, MSC_SERIAL, 0xf0);
 		}
 	}
@@ -464,31 +477,38 @@ static int wacom_intuos_inout(struct urb *urb)
 			default: /* Unknown tool */
 				wacom->tool[idx] = BTN_TOOL_PEN; break;
 		}
-		if(!((wacom->tool[idx] == BTN_TOOL_LENS) && 
-				(strstr(wacom->features->name, "Intuos3 12x12") 
-					|| strstr(wacom->features->name, "Intuos3 12x19"))))
-		{
-			input_report_abs(dev, ABS_MISC, wacom->id[idx]); /* report tool id */
-			input_report_key(dev, wacom->tool[idx], 1]);
-			input_event(dev, EV_MSC, MSC_SERIAL, wacom->serial[idx]);
-		}
 		return 1;
 	}
 
 	/* Exit report */
 	if ((data[1] & 0xfe) == 0x80)
 	{
+		input_report_abs(dev, ABS_X, 0);
+		input_report_abs(dev, ABS_Y, 0);
+		input_report_abs(dev, ABS_DISTANCE, 0);
+		if (wacom->tool[idx] >= BTN_TOOL_MOUSE) {
+			input_report_key(dev, BTN_LEFT, 0);
+			input_report_key(dev, BTN_MIDDLE, 0);
+			input_report_key(dev, BTN_RIGHT, 0);
+			input_report_key(dev, BTN_SIDE, 0);
+			input_report_key(dev, BTN_EXTRA, 0);
+			input_report_abs(dev, ABS_THROTTLE, 0);
+			input_report_abs(dev, ABS_RZ, 0);
+ 		} else {
+			input_report_abs(dev, ABS_PRESSURE, 0);
+			input_report_abs(dev, ABS_TILT_X, 0);
+			input_report_abs(dev, ABS_TILT_Y, 0);
+			input_report_key(dev, BTN_STYLUS, 0);
+			input_report_key(dev, BTN_STYLUS2, 0);
+			input_report_key(dev, BTN_TOUCH, 0);
+			input_report_abs(dev, ABS_WHEEL, 0);
+		}
 		input_report_key(dev, wacom->tool[idx], 0);
 		input_report_abs(dev, ABS_MISC, 0); /* reset tool id */
 		input_event(dev, EV_MSC, MSC_SERIAL, wacom->serial[idx]);
 		return 1;
 	}
-
-	if((wacom->tool[idx] == BTN_TOOL_LENS) && (strstr(wacom->features->name, "Intuos3 12x12") 
-			|| strstr(wacom->features->name, "Intuos3 12x19")))
-		return 1;
-	else
-		return 0;
+	return 0;
 }
 
 static void wacom_intuos_general(struct urb *urb)
@@ -562,10 +582,12 @@ static void wacom_intuos_irq(struct urb *urb)
 		input_report_abs(dev, ABS_RX, ((data[1] & 0x1f) << 8) | data[2]);
 		input_report_abs(dev, ABS_RY, ((data[3] & 0x1f) << 8) | data[4]);
 
-		if((data[5] & 0x0f) | (data[6] & 0x0f) | (data[1] & 0x1f) | data[2])
+		if((data[5] & 0x0f) | (data[6] & 0x0f) | (data[1] & 0x1f) | 
+			data[2] | (data[3] & 0x1f) | data[4])
 			input_report_key(dev, wacom->tool[1], 1);
 		else
 			input_report_key(dev, wacom->tool[1], 0);
+		input_report_abs(dev, ABS_MISC, PAD_DEVICE_ID);
 		input_event(dev, EV_MSC, MSC_SERIAL, 0xffffffff);
 		return;
 	}
@@ -573,6 +595,12 @@ static void wacom_intuos_irq(struct urb *urb)
 	/* process in/out prox events */
 	if (wacom_intuos_inout(urb)) return;
 
+	/* Only large I3 supports Lens Cursor */
+	if((wacom->tool[idx] == BTN_TOOL_LENS) && strstr(wacom->features->name, "Intuos3"))
+		if (!(strstr(wacom->features->name, "Intuos3 12x12") 
+			|| strstr(wacom->features->name, "Intuos3 12x19")))
+			return;
+
 	/* Cintiq doesn't send data when RDY bit isn't set */
 	if (strstr(wacom->features->name, "Cintiq") && !(data[1] & 0x40)) return;
 
@@ -667,15 +695,17 @@ static void wacom_intuos_irq(struct urb *urb)
 #define WACOM_GRAPHIRE_BITS	(BIT(EV_REL))
 #define WACOM_GRAPHIRE_REL	(BIT(REL_WHEEL))
 #define WACOM_G4_TOOLS		(BIT(BTN_TOOL_FINGER))
-#define WACOM_G4_BUTTONS	(BIT(BTN_0) | BIT(BTN_1) | BIT(BTN_2) | BIT(BTN_3) | BIT(BTN_4) | BIT(BTN_5) | BIT(BTN_6) | BIT(BTN_7))
+#define WACOM_G4_BUTTONS	(BIT(BTN_0) | BIT(BTN_4))
 #define WACOM_INTUOS_TOOLS	(BIT(BTN_TOOL_BRUSH) | BIT(BTN_TOOL_PENCIL) | BIT(BTN_TOOL_AIRBRUSH) | BIT(BTN_TOOL_LENS))
 #define WACOM_INTUOS3_TOOLS	(WACOM_INTUOS_TOOLS | BIT(BTN_TOOL_FINGER))
 #define WACOM_INTUOS_BUTTONS	(BIT(BTN_SIDE) | BIT(BTN_EXTRA))
-#define WACOM_INTUOS3_BUTTONS	(WACOM_INTUOS_BUTTONS | BIT(BTN_0) | BIT(BTN_1) | BIT(BTN_2) | BIT(BTN_3) | BIT(BTN_4) | BIT(BTN_5) | BIT(BTN_6) | BIT(BTN_7))
+#define WACOM_INTUOS3S_BUTTONS	(WACOM_INTUOS_BUTTONS | BIT(BTN_0) | BIT(BTN_1) | BIT(BTN_2) | BIT(BTN_3))
+#define WACOM_INTUOS3_BUTTONS	(WACOM_INTUOS3S_BUTTONS | BIT(BTN_4) | BIT(BTN_5) | BIT(BTN_6) | BIT(BTN_7))
 #define WACOM_INTUOS_BITS	(BIT(EV_REL))
 #define WACOM_INTUOS_REL	(BIT(REL_WHEEL))
 #define WACOM_INTUOS_ABS	(BIT(ABS_TILT_X) | BIT(ABS_TILT_Y) | BIT(ABS_RZ) | BIT(ABS_THROTTLE))
-#define WACOM_INTUOS3_ABS	(WACOM_INTUOS_ABS | BIT(ABS_RX) | BIT(ABS_RY))
+#define WACOM_INTUOS3S_ABS	(WACOM_INTUOS_ABS | BIT(ABS_RX))
+#define WACOM_INTUOS3_ABS	(WACOM_INTUOS3S_ABS | BIT(ABS_RY))
 
 struct wacom_features wacom_features[] = {
 
@@ -749,8 +779,8 @@ struct wacom_features wacom_features[] = {
 	/* 22 */ { "Wacom Intuos2 12x18", 10,  45720, 31680,  1023, 15,
 			wacom_intuos_irq, WACOM_INTUOS_BITS, WACOM_INTUOS_ABS,
 			WACOM_INTUOS_REL, WACOM_INTUOS_BUTTONS, WACOM_INTUOS_TOOLS },
-	/* Volito */
-	/* 23 */ { "Wacom Volito",         8,   5104,  3712,   511, 32,
+	/* Volito1 */
+	/* 23 */ { "Wacom Volito1",         8,   5104,  3712,   511, 32,
 			wacom_graphire_irq, WACOM_GRAPHIRE_BITS, 0, 0, 0 },
 	/* Volito2 - PenPartner - PenStation */
 	/* 24 */ { "Wacom PenStation2",    8,   3250,  2320,   255, 32,
@@ -775,8 +805,8 @@ struct wacom_features wacom_features[] = {
 			wacom_ptu_irq, 0, 0, 0, 0 },
 	/* Intuos3 */
 	/* 33 */ { "Wacom Intuos3 4x5",   10,  25400, 20320,  1023, 15,
-			wacom_intuos_irq, WACOM_INTUOS_BITS, WACOM_INTUOS3_ABS,
-			WACOM_INTUOS_REL, WACOM_INTUOS3_BUTTONS, WACOM_INTUOS3_TOOLS },
+			wacom_intuos_irq, WACOM_INTUOS_BITS, WACOM_INTUOS3S_ABS,
+			WACOM_INTUOS_REL, WACOM_INTUOS3S_BUTTONS, WACOM_INTUOS3_TOOLS },
 	/* 34 */ { "Wacom Intuos3 6x8",   10,  40640, 30480,  1023, 15,
 			wacom_intuos_irq, WACOM_INTUOS_BITS, WACOM_INTUOS3_ABS,
 			WACOM_INTUOS_REL, WACOM_INTUOS3_BUTTONS, WACOM_INTUOS3_TOOLS },
@@ -792,13 +822,16 @@ struct wacom_features wacom_features[] = {
 	/* 38 */ { "Wacom Intuos3 6x11",  10,  54204, 31750,  1023, 15,
 			wacom_intuos_irq, WACOM_INTUOS_BITS, WACOM_INTUOS3_ABS,
 			WACOM_INTUOS_REL, WACOM_INTUOS3_BUTTONS, WACOM_INTUOS3_TOOLS },
+	/* 39 */ { "Wacom Intuos3 4x6",   10,  31496, 19685,  1023, 15,
+			wacom_intuos_irq, WACOM_INTUOS_BITS, WACOM_INTUOS3S_ABS,
+			WACOM_INTUOS_REL, WACOM_INTUOS3S_BUTTONS, WACOM_INTUOS3_TOOLS },
 	/* Protocol 5 Cintiq */
-	/* 39 */ { "Wacom Cintiq 21UX",   10,  87200, 65600,  1023, 15,
+	/* 40 */ { "Wacom Cintiq 21UX",   10,  87200, 65600,  1023, 15,
 			wacom_intuos_irq, WACOM_INTUOS_BITS, WACOM_INTUOS3_ABS,
 			WACOM_INTUOS_REL, WACOM_INTUOS3_BUTTONS, WACOM_INTUOS3_TOOLS },
-	/* 40 */ { "Wacom DTF720",         8,   6858,  5506,   511, 32,
+	/* 41 */ { "Wacom DTF720",         8,   6858,  5506,   511, 32,
 			wacom_pl_irq, 0,  0, 0, 0 },
-	/* 41 */ { "Wacom DTF521",         8,   6282,  4762,   511, 32,
+	/* 42 */ { "Wacom DTF521",         8,   6282,  4762,   511, 32,
 			wacom_pl_irq, 0,  0, 0, 0 },
 
 	{ NULL , 0 }
@@ -844,9 +877,10 @@ struct usb_device_id wacom_ids[] = {
 	{ USB_DEVICE(USB_VENDOR_ID_WACOM, 0xB3), driver_info: 36 },
 	{ USB_DEVICE(USB_VENDOR_ID_WACOM, 0xB4), driver_info: 37 },
 	{ USB_DEVICE(USB_VENDOR_ID_WACOM, 0xB5), driver_info: 38 },
-	{ USB_DEVICE(USB_VENDOR_ID_WACOM, 0x3F), driver_info: 39 },
-	{ USB_DEVICE(USB_VENDOR_ID_WACOM, 0xC0), driver_info: 40 },
-	{ USB_DEVICE(USB_VENDOR_ID_WACOM, 0xC4), driver_info: 41 },
+	{ USB_DEVICE(USB_VENDOR_ID_WACOM, 0xB7), driver_info: 39 },
+	{ USB_DEVICE(USB_VENDOR_ID_WACOM, 0x3F), driver_info: 40 },
+	{ USB_DEVICE(USB_VENDOR_ID_WACOM, 0xC0), driver_info: 41 },
+	{ USB_DEVICE(USB_VENDOR_ID_WACOM, 0xC4), driver_info: 42 },
 
 	/* some Intuos2 6x8's erroneously report as 0x47;
 	 * multiple confirmed examples exist. */
@@ -933,8 +967,8 @@ static void *wacom_probe(struct usb_device *dev, unsigned int ifnum, const struc
 	wacom->dev.absmax[ABS_TILT_Y] = 127;
 	wacom->dev.absmax[ABS_WHEEL] = 1023;
 
-	wacom->dev.absmax[ABS_RX] = 4097;
-	wacom->dev.absmax[ABS_RY] = 4097;
+	wacom->dev.absmax[ABS_RX] = 4096;
+	wacom->dev.absmax[ABS_RY] = 4096;
 	wacom->dev.absmin[ABS_RZ] = -900;
 	wacom->dev.absmax[ABS_RZ] = 899;
 	wacom->dev.absmin[ABS_THROTTLE] = -1023;
